// Contains class definitions
// Copyright (C) 1997 by John P. Weiss
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the Artistic License, included as the file
// "LICENSE" in the source code archive.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the file "LICENSE", containing
// the License John Weiss originally placed this program under.
#include "WorkManDB.h"


// ***************
//  Here, we define the actual functions


// This function performs a ifstream.get() to get a token.  It checks
// that the delimiter isn't the first character on the stream, reads
// the token, then moves forward past the token.
//
bool getToken(ifstream& ifs, char *buf, int len, char delim)
{
   char junk;

   // Junk the delimiter.
   while(ifs.peek() == (int)delim)
	 ifs.get(junk);

   // Read the token.
   ifs.clear(0);
   ifs.get(buf,len,delim);
   if( (!ifs) || (ifs.gcount() == 0) )
	 {
	 cerr << "getToken() - read failure." << endl;
	 return false;
	 }

   // Junk the delimiter.
   while(ifs.peek() == (int)delim)
	 ifs.get(junk);

   return true;
}


//||||||||||||||||||||||||||||||||||||||||||||||||
//
//  Class WorkManDB
//
//||||||||||||||||||||||||||||||||||||||||||||||||



// Constructor for the WorkManDB object.  Parent AudDB(void) called
// first. 
//
WorkManDB::WorkManDB(void)
{
   seeklist = NULL;
}


// Overloaded constructor for the WorkManDB object.  Parent
// AudDB(void) called first.
//
WorkManDB::WorkManDB(char *fname) : AudDB(fname)
{
   seeklist = NULL;
   open();
}


// Destructor for the WorkManDB object.
//
WorkManDB::~WorkManDB(void)
{
   if(seeklist != NULL)
	 delete [] seeklist;
}


// Opens the catalog and reads in, amongst other things, magic numbers
// and finds seek-positions.
//
void WorkManDB::open(void)
{
   char buf[IOBUFLEN];
   ifstream ifs;

   ifs.open( db_name, ios::in | ios::nocreate | ios::binary );
   if (!ifs)
	 {
	 cerr << "WorkmanDB: Cannot open database, '"
		 << db_name << "'." << endl;
	 return;
	 }

   // First pass - counts the number of records for array allocation
   // purposes.
   len_magic=0;
   while( ifs.good() )
	 {
	 ifs.getline(buf,IOBUFLEN);
	 if(strstr(buf,"tracks ") != NULL)
	    len_magic++;
	 }
   ifs.clear(0);
   ifs.seekg(0);
   if(len_magic == 0)
	 {
	 cerr << "WorkmanDB: Empty or corrupt database, '"
		 << db_name << "'." << endl;
	 ifs.close();
	 return;
	 }

   // Second pass - call rescanDB()
   rescanDB(ifs);
   ifs.close();
}


// Scans the database record by record.  Gets the magic numbers and
// record positions.
//
void WorkManDB::rescanDB(ifstream& ifs)
{  
   int i;
   bool okay;
   Catalog wrkbuf;

   if(magiclist != NULL)
	 delete [] magiclist;
   if(seeklist != NULL)
	 delete [] seeklist;
   magiclist = new unsigned long[len_magic];
   seeklist = new streampos[len_magic];
   for(i=0; i<len_magic; i++)
	 {
	 seeklist[i] = ifs.tellg();
	 okay = readEntry(ifs,wrkbuf);
	 if(okay)
	    {
	    magiclist[i] = wrkbuf.magic();
	    }
	 else
	    {
	    magiclist[i] = 0L;
	    seeklist[i] = 0;
	    }
	 }
}


// Reads catalog entry #idx.
//
bool WorkManDB::read_i(Catalog& ct, int idx)
{
   bool stat;
   ifstream ifs;

   if( (idx < 0) || (idx >= len_magic) )
	 return false;

   ifs.open( db_name, ios::in | ios::nocreate | ios::binary );
   if (!ifs)
	 {
	 cerr << "WorkmanDB::read() - Cannot open database, '"
		 << db_name << "'." << endl;
	 return false;
	 }
   ifs.seekg(seeklist[idx]);
   stat = readEntry(ifs,ct);
   ifs.close();
   return stat;
}


// Writes or overwrites an entry into the catalog.
//
void WorkManDB::write(Catalog& ct, int& oldlen)
{
   bool stat;
   int idx;

   idx = find(ct);
   if(idx < 0)
	 {
	 stat = appendEntry(ct);
	 if(stat)
	    oldlen++;
	 }
   else
	 {
	 stat = replaceEntry(ct);
	 }
}


// Method to read entire catalog.  Performs no file seeking.
//
bool WorkManDB::readEntry(ifstream& ifs, Catalog& ct)
{
   int nucatlen,i,wrktime;
   char buf[IOBUFLEN];
   Catalog *resizer_;
   const char *header="WorkManDB::readEntry() - ";
   const char *badrec="bad record.";
   const char *readerr="read error.";


   // First, seek the track information
   getToken(ifs,buf,IOBUFLEN,' ');
   if(strcmp(buf,"tracks") != 0)
	 {
	 cerr << header << badrec << endl;
	 cerr << "Record should begin with 'tracks' entry, not '"
		 << buf << "'." << endl;
	 return false;
	 }

   ifs >> nucatlen;
   if( (nucatlen <= 0) || (!ifs) )
	 {
	 cerr << header << readerr << endl;
	 return false;
	 }
	 
   // Resize the old catalog.
   resizer_ = new Catalog(nucatlen);
   ct = (*resizer_);
   delete resizer_;

   // Get the track time offset info.  The device also re-reads this
   // info.
   for(i=0; i<nucatlen+1; i++)
	 {
	 ifs >> wrktime;
	 if( (wrktime < 0) || (!ifs) )
	    {
	    cerr << header << readerr << endl;
	    return false;
	    }
	 ct.setTime(i,wrktime);
	 }
   ifs.ignore(IGNORECOUNT,'\n');

   // Now get the title and artist.
   getToken(ifs,buf,IOBUFLEN,' ');
   if(strcmp(buf,"cdname") != 0)
	 {
	 cerr << header << badrec << endl;
	 cerr << "The 'cdname' entry comes next, not '"
		 << buf << "'." << endl;
	 return false;
	 }
   getToken(ifs,buf,MAXCHARLEN,'\n');
   ct.setTitle(buf);

   getToken(ifs,buf,IOBUFLEN,' ');
   if(strcmp(buf,"artist") != 0)
	 {
	 cerr << header << badrec << endl;
	 cerr << "The 'artist' entry comes next, not '"
		 << buf << "'." << endl;
	 return false;
	 }
   getToken(ifs,buf,MAXCHARLEN,'\n');
   ct.setArtist(buf);
   
   // Last, read in the track names.
   for(i=0; i<nucatlen; i++)
	 {
	 getToken(ifs,buf,IOBUFLEN,' ');
	 if(strcmp(buf,"track") != 0)
	    {
	    cerr << header << badrec << endl;
	    cerr << "Read '" << buf << "' instead of 'track' entry." 
		    << endl;
	    return false;
	    }
	 getToken(ifs,buf,IOBUFLEN,'\n');
	 ct.setName(i,buf);
	 }

   // We made it!  Read successful!
   return true;
}


// Method to read entire catalog entry.  Performs no file seeking.
// Unlike the earlier version, this one also alters the track data
// list instead of just reading and ignoring it.
//
bool readEntry_all_WorkManDB(ifstream& ifs, Catalog& ct)
{
   int nucatlen,i,wrktime;
   char buf[IOBUFLEN];
   Catalog *working=NULL;
   const char *header="WorkManDB::readEntry() - ";
   const char *badrec="bad record.";
   const char *readerr="read error.";


   // First, seek the track information
   getToken(ifs,buf,IOBUFLEN,' ');
   if(strcmp(buf,"tracks") != 0)
	 {
	 cerr << header << badrec << endl;
	 cerr << "Record should begin with 'tracks' entry, not '"
		 << buf << "'." << endl;
	 return false;
	 }

   ifs >> nucatlen;
   if( (nucatlen <= 0) || (!ifs) )
	 {
	 cerr << header << readerr << endl;
	 return false;
	 }
	 
   working = new Catalog(nucatlen);
   for(i=0; i<nucatlen+1; i++)
	 {
	 ifs >> wrktime;
	 if( (wrktime < 0) || (!ifs) )
	    {
	    cerr << header << readerr << endl;
	    return false;
	    }
	 working->setTime(i,wrktime);
	 }
   ifs.ignore(IGNORECOUNT,'\n');

   // Now get the title and artist.
   getToken(ifs,buf,IOBUFLEN,' ');
   if(strcmp(buf,"cdname") != 0)
	 {
	 cerr << header << badrec << endl;
	 cerr << "The 'cdname' entry comes next, not '"
		 << buf << "'." << endl;
	 return false;
	 }
   getToken(ifs,buf,MAXCHARLEN,'\n');
   working->setTitle(buf);

   getToken(ifs,buf,IOBUFLEN,' ');
   if(strcmp(buf,"artist") != 0)
	 {
	 cerr << header << badrec << endl;
	 cerr << "The 'artist' entry comes next, not '"
		 << buf << "'." << endl;
	 return false;
	 }
   getToken(ifs,buf,MAXCHARLEN,'\n');
   working->setArtist(buf);
   
   // Last, read in the track names.
   for(i=0; i<nucatlen; i++)
	 {
	 getToken(ifs,buf,IOBUFLEN,' ');
	 if(strcmp(buf,"track") != 0)
	    {
	    cerr << header << badrec << endl;
	    cerr << "Read '" << buf << "' instead of 'track' entry." 
		    << endl;
	    return false;
	    }
	 getToken(ifs,buf,IOBUFLEN,'\n');
	 working->setName(i,buf);
	 }

   // We made it!  Read successful!
   working->genMagic();
   ct = (*working);
   delete working;
   return true;
}


// Method to write entire catalog.  Performs no file seeking.
//
bool WorkManDB::writeEntry(ofstream& ofs, Catalog& ct)
{
   bool badness;
   int i,clen;

   clen = ct.len();
   badness = false;
   ofs << "tracks " << clen;
   badness |= (!ofs);
   // Don't forget the leadout track!
   for(i=0; i<=clen; i++)
	 {
	 ofs << " " << ct.time(i);
	 badness |= (!ofs);
	 }
   ofs << endl;
   badness |= (!ofs);

   // Make sure there are no empty strings.  The read routine gags on
   // empty fields.
   if (strlen(ct.title()) == 0)
	 ofs << "cdname NoTitle" << endl;
   else
	 ofs << "cdname " << ct.title() << endl;
   badness |= (!ofs);
   if (strlen(ct.artist()) == 0)
	 ofs << "artist -" << endl;
   else
	 ofs << "artist " << ct.artist() << endl;
   badness |= (!ofs);
   for(i=0; i<clen; i++)
	 {
	 if (strlen(ct.name(i)) == 0)
	    ofs << "track " << i+1 << endl;
	 else
	    ofs << "track " << ct.name(i) << endl;
	 badness |= (!ofs);
	 }
   
   if(badness)
	 {
	 cerr << "WorkManDB::writeEntry() - write failure."
		 << "Entry mangled." << endl;
	 cerr << "[You may be able to fix the problem by editing '"
		 << db_name << "' by hand.]" << endl;	 
	 return false;
	 }

   return true;
}


// Method to delete an entry in the catalog.  Don't be intimidated by
// the length of the code here - it's mostly error checking to be
// EXTRA CAREFUL that we don't lose anything from the database.
//
bool WorkManDB::deleteEntry(Catalog& ct)
{
   int idx, bloclen;
   ifstream ifs;
   ofstream ofs;
   streampos flen;
   char *readbuf;
   const char *header="WorkManDB::deleteEntry() - ";
   const char *seekerr="seek error.  ";
   const char *safeabort="No delete performed.";
   const char 
	 *ocsrmesg="Old database lost.  A copy still resides in '";


   idx = find(ct);
   if(idx < 0)
	 {
	 cerr << header << "bizarre mystery error."  << safeabort
		 << endl;
	 return false;
	 }

   ifs.open( db_name, ios::in | ios::nocreate | ios::binary );
   if (!ifs)
	 {
	 cerr << header << "Cannot open database, '" << db_name 
		 << "'." << endl;
	 cerr << safeabort << endl;
	 return false;
	 }
   ofs.open( TMPDBFILE, ios::out | ios::trunc | ios::binary );
   if (!ofs)
	 {
	 cerr << header << "Cannot open temporary storage file."
		 << endl;
	 cerr << safeabort << endl;
	 return false;
	 }

   // After opening, we copy everything before the old entry into the
   // temp-file.
   bloclen = (int)seeklist[idx];
   readbuf = new char[bloclen];
   ifs.read(readbuf,bloclen);
   if (!ifs)
	 {
	 cerr << header << "read error.  " << safeabort << endl;
	 ifs.close();
	 ofs.close();
	 delete [] readbuf;
	 return false;
	 }
   ofs.write(readbuf,bloclen);
   if (!ofs)
	 {
	 cerr << header << "write error.  " << safeabort << endl;
	 ifs.close();
	 ofs.close();
	 delete [] readbuf;
	 return false;
	 }
   delete [] readbuf;

   // If there's stuff after the target record, append it, too.
   if(idx+1 < len_magic)
	 {
	 ifs.seekg(0,ios::end);
	 flen = ifs.tellg();
	 bloclen = (int)(flen-seeklist[idx+1]);
	 ifs.seekg(seeklist[idx+1]);
	 if( (flen <= 0) || (bloclen <= 0) || (!ifs) )
	    {
	    cerr << header << seekerr << safeabort << endl;
	    ifs.close();
	    ofs.close();
	    return false;
	    }

	 readbuf = new char[bloclen];
	 ifs.read(readbuf,bloclen);
	 if (!ifs)
	    {
	    cerr << header << "read error.  " << safeabort << endl;
	    ifs.close();
	    ofs.close();
	    delete [] readbuf;
	    return false;
	    }
	 ofs.write(readbuf,bloclen);
	 if (!ofs)
	    {
	    cerr << header << "write error.  " << safeabort << endl;
	    ifs.close();
	    ofs.close();
	    delete [] readbuf;
	    return false;
	    }
	 delete [] readbuf;
	 }
   ofs.close();
   ifs.close();

   // Now we copy the temp file to the old one.
   // Do the reading first.  Also rescan the database from this file
   // before performing the copy.
   ifs.open( TMPDBFILE, ios::in | ios::nocreate | ios::binary );
   if (!ifs)
	 {
	 cerr << header << "Cannot open temporary storage file."
		 << endl;
	 cerr << safeabort << endl;
	 return false;
	 }
   ifs.seekg(0,ios::end);
   flen = ifs.tellg();
   bloclen = (int)flen;
   ifs.seekg(0);
   if( (flen <= 0) || (bloclen <= 0) || (!ifs) )
	 {
	 cerr << header << seekerr << safeabort << endl;
	 ifs.close();
	 return false;
	 }
   readbuf = new char[bloclen];
   ifs.read(readbuf,bloclen);
   if (!ifs)
	 {
	 cerr << header << "read error.  " << safeabort << endl;
	 ifs.close();
	 delete [] readbuf;
	 return false;
	 }
   ifs.seekg(0);
   len_magic--;
   rescanDB(ifs);
   ifs.close();

   // Open the file and write - this is the dangerous part!
   ofs.open( db_name, ios::out | ios::trunc | ios::binary );
   if (!ofs)
	 {
	 cerr << header << "Fatal error opening database, '" << db_name 
		 << "'." << endl;
	 cerr << ocsrmesg << TMPDBFILE << "'." << endl;
	 return false;
	 }
   ofs.write(readbuf,bloclen);
   if (!ofs)
	 {
	 cerr << header << "Fatal write error.  " << safeabort << endl;
	 cerr << ocsrmesg << TMPDBFILE << "'." << endl;
	 ofs.close();
	 delete [] readbuf;
	 return false;
	 }
   delete [] readbuf;
   ofs.close();

   // Phew!  We made it.  Now rescan the database.
   return true;
}


// Method to append a catalog to the end.
//
bool WorkManDB::appendEntry(Catalog& ct)
{
   bool stat;
   int i;
   ofstream ofs;
   streampos *old,locn;

   ofs.open( db_name, ios::in | ios::out | ios::app | ios::ate |
		  ios::binary ); 
   if (!ofs)
	 {
	 cerr << "WorkmanDB::appendEntry() -  Cannot open database, '"
		 << db_name << "'." << endl;
	 return false;
	 }

   locn = ofs.tellp();
   stat = writeEntry(ofs,ct);
   ofs.close();
   if(!stat)
	 return false;

   old = seeklist;
   seeklist = new streampos[len_magic+1];
   for(i=0; i<len_magic; i++)
	 seeklist[i] = old[i];
   seeklist[len_magic] = locn;
   if(old != NULL)
	 delete [] old;

   return true;
}


// Method to replace an entry in the catalog.
//
bool WorkManDB::replaceEntry(Catalog& ct)
{
   bool stat;
   unsigned long *oldml;
   int i;
   const char *header="WorkmanDB::replaceEntry() - operation failed.";

   stat = deleteEntry(ct);
   if(!stat)
	 {
	 cerr << header << endl;
	 cerr << "[You may be able to fix the problem by editing '"
		 << db_name << "' by hand.]" << endl;	 
	 return false;
	 }
   stat = appendEntry(ct);
   if(!stat)
	 {
	 cerr << header << "  New entry lost." << endl;	 
	 return false;
	 }
   
   // Lastly, we re-add the magic number since:
   //   1) write won't change anything, since we've not added any
   //   entries but replaced and existing one.
   //   2) deleteEntry() removes the old entries
   //   3) appendEntry() adds the new scan position to the list but
   //   doesn't change anything else.
   oldml = magiclist;
   magiclist = new unsigned long[len_magic+1];
   for(i=0; i<len_magic; i++)
	 magiclist[i] = oldml[i];
   magiclist[len_magic] = ct.magic();
   len_magic++;
   if(oldml != NULL)
	 delete [] oldml;
   return true;
}







//||||||||||||||||||||||||||||||||
//
// End
//
