//-*-C++-*- Header file for the AudioCD Class
// Copyright (C) 1997 by John P. Weiss
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the Artistic License, included as the file
// "LICENSE" in the source code archive.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the file "LICENSE", containing
// the License John Weiss originally placed this program under.
//
// This file is unreadable, because the comment-docifier is
// brainedead.  Try a "make srcdocs" to generate the
// HTML-documentation.
//
#ifndef _AudioCD_h_
#define _AudioCD_h_
#include <qtimer.h>
#include "RawCD.h"
#include "AudioDev.h"
//


// These next five are in units of msec.
#define UPDATE_INTERVAL 500
#define DT_PLAYNEXT_NOW 50
#define SCANRATE_INCR_INTERVAL 5000
#define INITIAL_SCAN_STEP 250
#define FASTEST_SCAN_STEP 15




/**
  The {\bf AudioCD} class is the interface to a \Ref{RawCD} object.
  It is toolkit-dependent and also provides an interface to the
  \Ref{AudioPanel} object via inheritance of the \Ref{AudioDev}
  class.
    
  Repeating a single item, an entire playlist, or shuffling a
  playlist are all intrinsic parts of this class.  All of these
  functions operate on the current playlist, as opposed to the entire
  CD.

  All public methods, slots and signals are abstract virtual members
  of the \Ref{AudioDev} class.  They've been implemented as described
  in the documentation of \Ref{AudioDev}, but with support for
  programs, shuffling, and repeating play.  I've only documented those
  with greater functionality than what's described in the
  documentation of \Ref{AudioDev}.

  The private members [which includes the \Ref{RawCD} object] and
  methods are not documented, as they exist solely to support the
  implementations of all of the public methods, slots, and signals.
  One of the neat things about the private members are slots that
  receive signals from Qt-timers.  Using the timers to keep track of
  the elapsed time, I can support even the most buggy of CDROM drives!

  {\bf
  \begin{verbatim}
  
  ----------------------------------------------------------
  ----------------------------------------------------------
  \end{verbatim}
  }

  @memo The Qt-interface to a \Ref{RawCD} object.
  @see RawCD, Catalog, AudioDev

  */
class AudioCD : public AudioDev
{
	 //
	 Q_OBJECT

   public:
	 /**
	   Class Constructor
	   @memo
	   */
	 AudioCD( void );

	 /** {\tt [virtual]}

	   Reimplemented from \Ref{AudioDev}.  The {\bf AudioCD} object
	   remains empty until this is called.  This implementation {\em
	   could} be called multiple times.
	   @memo
	  */
	 void DeviceFirstInit( char *devname );

	 /** 
	   Wrapper for the \Ref{RawCD} version.
	   @memo
	  */
	 void TrayControl( bool close_ioctl=false, 
				    bool trayless=false ) {
	    player.TrayControl(close_ioctl,trayless);
	 }

	 //
	 void newProgram(void);

	 /** {\tt [virtual]}

	   Reimplemented from \Ref{AudioDev}.  Emits signals
	   to update the title, artist, and current item label
	   [\Ref{newProgLabel}, \Ref{newProgArtist}, \Ref{newItemLabel}.
	   Does not emit \Ref{newTime} or \Ref{newItemNumber}, however.	   
	   @memo
	   */
	 void cleanDisplay(void) {
	    char *tmp;
	    tmp = ToC->plName((unsigned int)currentidx);
	    emit newItemLabel(tmp);
	    tmp = (char *)(ToC->title());
	    emit newProgLabel(tmp);
	    showArtist();
	 }

	 /** {\tt [virtual]}

	   Reimplemented from \Ref{AudioDev}.  Scans the CD player to
	   update the track list in the ToC.  This is here so that
	   programs that read a catalog entry from a database can signal
	   the device to do this sort of update.  
	   @memo
	   */
	 void updateToCTrkInf(void) {
	      int i, l;
		 l = ToC->len();
		 ToC->setTime(l,player.ToC_Entry(l));
		 for(i=0, proglen=0; i<l; i++)
		    {
		    ToC->setTime(i,player.ToC_Entry(i));
		    itemLengths[i] = player.ItemLength(ToC->playlist[i]);
		    proglen += itemLengths[i];
		    }
	 }
	 
	 
   public slots:
	 //
	 void StartScanFwd();
	 void StartScanRev();
	 void StopScan();
	 void StartPlayOrPause();

	 /**
	   {\em Public Slot} {\tt [virtual]}
	   
	   Reimplemented from \Ref{AudioDev}.  Stops play and resets the
	   CDROM device [via \Ref{RawCD}].  The first call also
	   unshuffles the playlist, but does not erase any special
	   program.  A second call after the first resets the playlist,
	   erasing any custom program. 
	   @memo
	   */
	 void StopPlay();
	 //
	 void Next();
	 void Prev();
	 void IncVol();
	 void DecVol();
	 void newTimeMode(int);
	 void newPlayMode(int);
	 
	 /**
	   {\em Public Slot} {\tt [virtual]}
	   
	   Reimplemented from \Ref{AudioDev} to eject a CD or retract the
	   tray.	   
	   @memo
	   */
	 void LoadOrEject();

   private:
	 //
      RawCD player;
	 int proglen, *itemLengths;
	 int *deshuffle_buf;
	 int currentidx, nextidx;
	 int scantime, scanstep;
	 bool empty, do_clear_all;
	 bool scanning_forward, scan_not_playing;
	 bool repeat1, repeat_all, shuffle;
	 bool show_total, show_remain;
	 bool at_end, do_repause;
	 //
	 QTimer *scanrateT, *scanT, *updateT, *playnextT;
	 //
	 void auInitToC(void);
	 void auShuffle(void);
	 void auResyncProgram(void);
	 inline void auUnshuffle(void);
	 inline void auInitTimers(void);
	 inline void auTogglePause(void);
	 inline void auResetFlags(void);
	 inline void auEmptyToC(void);
	 inline void auResetPlayer(bool checkForEmpty=true);
	 inline void auShowTime(int nutime);
	 inline void auShowIndex(int nuidx);
	 inline void auShowEmpty(void);
	 inline int auPI_Len(int idx);
	 inline void auUpdateProgLen(void);
	 void showArtist(void) {
	    if(ToC->artist() == (char *)0)
		  emit newProgArtist(0);		  
	    if((ToC->artist())[0] == 0)
		  emit newProgArtist(0);
	    emit newProgArtist( (char *)(ToC->artist()) );
	 }

   private slots:
	 //
	 void auDoScan();
	 void auScanRate();
      void auPlayNext();
	 void auUpdateTime();
};


#endif // _AudioCD_h_
//||||||||||||||||||||||||||||||||
//
// End
// 
